"""Top-level package for pyfusion."""

__author__ = "Fusion Devs"
__email__ = "fusion_developers@jpmorgan.com"
__version__ = "1.0.0"

from .fusion import Fusion  # Import the core Fusion class

__all__ = ["Fusion"]  # Explicitly define public API
